﻿# coding:utf-8
# --------------------------------------------------------------------
# Name:       心射円筒図法・メルカトル図法・正射円筒図法（仮称）の作成
# Created:    2019年4月6日
# Copyright:  WINGFIELD since1981
# 説明:       地理座標系のフィーチャクラスを投影座標系の座標値に換算
#             球体のみサポート
#             XY ドメインの範囲外に出力される場合はエラー（シェープファイル出力推奨）
#             Null ジオメトリをが含まれているとエラー
#             エラーが発生するとフィーチャクラスがロックされる（要再起動）  
#
# 参考
# http://w3e.kanazawa-it.ac.jp/e-scimath/contents/t15/textbook_t15_all.pdf
# https://ja.wikipedia.org/wiki/%E3%83%A1%E3%83%AB%E3%82%AB%E3%83%88%E3%83%AB%E5%9B%B3%E6%B3%95
# https://ja.wikipedia.org/wiki/%E3%82%B0%E3%83%BC%E3%83%87%E3%83%AB%E3%83%9E%E3%83%B3%E9%96%A2%E6%95%B0
# http://mogist.kkc.co.jp/word/0d28a32e-2164-45e6-9bef-920ffb1ef084.html
# --------------------------------------------------------------------

import arcpy
import math
import sys

# 環境設定
arcpy.env.overwriteOutput = True

try:
  # --------------------------------------------------------------------
  # 処理
  # --------------------------------------------------------------------
  # 入力パラメーター
  input_features = arcpy.GetParameterAsText(0)        # 入力フィーチャ
  output_featureclass = arcpy.GetParameterAsText(1)   # 出力フィーチャクラス
  output_projection = arcpy.GetParameterAsText(2)     # 投影法名称（'CPC' MERCATOR' 'MILLER' 'LCEA' 'PC'）
  radius = arcpy.GetParameterAsText(3)                # 地球半径（未入力で赤道長を 360 に設定）
  lat_limit = arcpy.GetParameterAsText(4)             # 限界緯度（メルカトル図法・心射円筒図法）

  # パラメーター確認
  if input_features == None or input_features == '':
    input_features = r'D:\temp\crs\GCS_LatLonLine.shp'
  
  if output_featureclass == None or output_featureclass == '':
    output_featureclass = r'D:\temp\crs\_project.shp'

  if output_projection == None or output_projection == '':
    output_projection = 'PC'

  if radius == None or radius == '':
    radius = 6378137.0 #180 / math.pi  # 赤道長を 360 とした設定 WGS84 の場合は 6378137.0
  else:
    radius = float(radius)
  
  if lat_limit == None or lat_limit == '':
    lat_limit = 89.0
  else:
    lat_limit = float(lat_limit)




  # --------------------------------------------------------------------
  # 関数
  # 半径（地理座標系：180/math.pi、WGS84赤道半径：6378137）
  # --------------------------------------------------------------------
  # 地理座標系を心射円筒図法に換算
  def get_central_perspective_cylindrical(lon, lat, radius, lat_limit):
    # 限界緯度の設定
    if lat > lat_limit:
      lat = lat_limit
    if lat < -(lat_limit):
      lat = -(lat_limit)
      
    x = lon / math.degrees(1) * radius
    y = math.tan(math.radians(lat)) * radius
    return [x, y]

  # 地理座標系をメルカトル座標系に換算
  def get_mercator(lon, lat, radius, lat_limit):
    # 限界緯度の設定
    if lat > lat_limit:
      lat = lat_limit
    if lat < -(lat_limit):
      lat = -(lat_limit)
    
    x = lon / math.degrees(1) * radius
    y = math.atanh(math.sin(math.radians(lat))) * radius  # グーデルマン関数
    return [x, y]

  # 地理座標系をミラー図法に換算
  def get_miller(lon, lat, radius):
    x = lon / math.degrees(1) * radius
    y = math.atanh(math.sin(math.radians(lat* 0.8))) * radius / 0.8
    return [x, y]

  # 地理座標系を正距円筒図法（標準緯線 0度）に換算
  def get_plate_carree(lon, lat, radius):
    x = lon / math.degrees(1) * radius
    y = lat / math.degrees(1) * radius
    return [x, y]

  # 地理座標系をランベルト正積円筒図法（正射円筒図法）に換算
  def get_lambert_cyrindrical_equal_area(lon, lat, radius):    
    x = lon / math.degrees(1) * radius
    y = math.sin(math.radians(lat)) * radius
    return [x, y]

  # 投影法の選択
  def get_projection(lon, lat, radisus, lat_limit, output_projection):
    if output_projection == 'CPC':
      return get_central_perspective_cylindrical(lon, lat, radius, lat_limit)
    elif output_projection == 'MERCATOR':
      return get_mercator(lon, lat, radius, lat_limit)
    elif output_projection == 'MILLER':
      return get_miller(lon, lat, radius)
    elif output_projection == 'PC':
      return get_plate_carree(lon, lat, radius)
    elif output_projection == 'LCEA':
      return get_lambert_cyrindrical_equal_area(lon, lat, radius)




  # --------------------------------------
  # 実行
  # --------------------------------------
  arcpy.AddMessage('処理開始...')
  print('処理開始...')

  # フィーチャ タイプ・ジオメトリ タイプの取得
  describe_featureclass = arcpy.Describe(input_features)
  feature_type = describe_featureclass.featureType
  shape_type = describe_featureclass.shapeType

  # 前提条件の確認
  if feature_type != 'Simple':
    arcpy.AddError('コンポジット フィーチャは処理できません。')
    print('コンポジット フィーチャは処理できません。')
    sys.exit()

  if hasattr(describe_featureclass.spatialReference, "PCS"):
    arcpy.AddError('投影座標系は処理できません。')
    print('投影座標系は処理できません。')
    sys.exit()

  # フィーチャのコピー
  temp_output = arcpy.CopyFeatures_management(input_features, output_featureclass).getOutput(0)

  # 空間参照の削除
  # 未実装（ArcGIS Pro 2.2 時点では Unknown Coordinate System は作成できない）

  # レコード数のカウント
  record_count = int(arcpy.GetCount_management(input_features).getOutput(0))

  # Progressor の設定
  arcpy.ResetProgressor()
  arcpy.SetProgressor('step', 'フィーチャを処理しています...', 1, record_count, 1)

  # UpdateCursor 関数を使用して Cursorオブジェクトを取得
  cursor = arcpy.da.UpdateCursor(temp_output, 'SHAPE@')

  # ループ カウンター
  counter = 1

  # cursor 内の row オブジェクトを使用してフィーチャクラスが持つ SHAPE フィールドの値を更新
  for row in cursor:
    # ループ カウンター
    arcpy.AddMessage('{0}/{1} フィーチャ処理...'.format(counter, record_count))
    print('{0}/{1} フィーチャ処理...'.format(counter, record_count))

    geometry = row[0]

    # ジオメトリ タイプ：ポイント
    if shape_type == 'Point':
      point = geometry.getPart(0)
      point.X = get_projection(point.X , point.Y, radius, lat_limit, output_projection)[0]
      point.Y = get_projection(point.X, point.Y, radius, lat_limit, output_projection)[1]
      geometry = arcpy.Point(point.X, point.Y)

    # ジオメトリ タイプ：マルチポイント
    elif shape_type == 'Multipoint':
      newArray = arcpy.Array()

      for idx in range(geometry.partCount):
        point = geometry.getPart(idx)

        point.X = get_projection(point.X , point.Y, radius, lat_limit, output_projection)[0]
        point.Y = get_projection(point.X , point.Y, radius, lat_limit, output_projection)[1]
        newArray.add(point)

      geometry = arcpy.Multipoint(newArray)

    # ジオメトリ タイプ：ポリライン、ポリゴン
    elif shape_type == 'Polyline' or shape_type == 'Polygon':
      newArray = arcpy.Array()

      for idx in range(geometry.partCount):
        part = geometry.getPart(idx)

        newPartArray = arcpy.Array()

        # part(Array 型) が None になるまでループ
        part.reset()
        point = part.next()

        while point != None:
          point.X = get_projection(point.X , point.Y, radius, lat_limit, output_projection)[0]
          point.Y = get_projection(point.X , point.Y, radius, lat_limit, output_projection)[1]
          newPartArray.add(point)

          point = part.next()

        newArray.add(newPartArray)

      # ジオメトリ タイプ：ポリライン
      if shape_type == 'Polyline':
        geometry = arcpy.Polyline(newArray)
      # ジオメトリ タイプ：ポリゴン
      elif shape_type == 'Polygon':
        geometry = arcpy.Polygon(newArray)

    # ジオメトリ タイプ：マルチパッチ
    else:
      arcpy.AddError('入力フィーチャクラスのジオメトリ タイプは処理できません。')
      sys.exit()

    # 更新したジオメトリのセット
    row[0] = geometry

    # フィーチャのアップデート
    cursor.updateRow(row)

    # ステップ
    arcpy.SetProgressorPosition()

    # ループ カウンター
    counter += 1

  # オブジェクト参照を解放
  del row, cursor, temp_output, input_features, output_featureclass

  arcpy.AddMessage('{0} フィーチャを処理しました。'.format(str(record_count)))
  print('{0} フィーチャを処理しました。'.format(str(record_count)))

except arcpy.ExecuteError as e:
  arcpy.AddError('エラーが発生しました。')
  arcpy.AddError(str(e))
  print('エラーが発生しました。')
  print(str(e))

  # オブジェクト参照を解放
  del row, cursor, temp_output, input_features, output_featureclass
